# coding: utf-8
from collections import OrderedDict
from datetime import datetime
import io
import json
import logging
from math import ceil
import os
import pickle
import re
import traceback

from flask import Blueprint, redirect, render_template, request, send_file, url_for

from ...models import Task, db
from ...vars import RUN_SPIDER_HISTORY_LOG, UA_DICT
from ..baseview import BaseView
from .execute_task import execute_task
from .utils import slot


apscheduler_logger = logging.getLogger('apscheduler')


def generate_cmd(auth, url, data):
    if auth:
        cmd = 'curl -u %s:%s %s' % (auth[0], auth[1], url)
    else:
        cmd = 'curl %s' % url

    for key, value in data.items():
        if key == 'setting':
            for v in value:
                t = (tuple(v.split('=', 1)))
                if v.startswith('USER_AGENT='):
                    cmd += ' --data-urlencode "setting=%s=%s"' % t
                else:
                    cmd += ' -d setting=%s=%s' % t
        elif key != '__task_data':
            cmd += ' -d %s=%s' % (key, value)

    return cmd


bp = Blueprint('schedule', __name__, url_prefix='/')


@bp.route('/schedule/history/')
def history():
    return send_file(RUN_SPIDER_HISTORY_LOG, mimetype='text/plain', cache_timeout=0)


class ScheduleView(BaseView):

    def __init__(self):
        super(ScheduleView, self).__init__()

        self.project = self.view_args['project']
        self.version = self.view_args['version']
        self.spider = self.view_args['spider']
        self.task_id = request.args.get('task_id', default=None, type=int)
        self.task = None

        self.url = 'http://%s/schedule.json' % self.SCRAPYD_SERVER
        self.template = 'scrapydweb/schedule.html'
        self.kwargs = {}

        self.selected_nodes = []
        self.first_selected_node = None

    def dispatch_request(self, **kwargs):
        if self.task_id:
            self.task = Task.query.get(self.task_id)
            if not self.task:
                message = "Task #%s not found" % self.task_id
                self.logger.error(message)
                return render_template(self.template_fail, node=self.node, message=message)
            self.query_task()
        elif self.POST:
            self.selected_nodes = self.get_selected_nodes()
            self.first_selected_node = self.selected_nodes[0]
        else:
            if self.project:
                # START button of Jobs page / Run Spider button of Logs page
                self.selected_nodes = [self.node]
            else:
                self.selected_nodes = []
            self.first_selected_node = self.node

        self.update_kwargs()
        return render_template(self.template, **self.kwargs)

    def query_task(self):
        task = self.task
        self.project = task.project
        self.version = task.version
        self.spider = task.spider

        self.selected_nodes = json.loads(task.selected_nodes)
        self.first_selected_node = self.selected_nodes[0]

        # 'settings_arguments': {'arg1': '233', 'setting': ['CLOSESPIDER_PAGECOUNT=10',]}
        settings_arguments = json.loads(task.settings_arguments)
        self.kwargs['expand_settings_arguments'] = len(settings_arguments) > 1 or settings_arguments['setting']
        settings_dict = dict(s.split('=') for s in settings_arguments.pop('setting'))
        arguments_dict = settings_arguments

        self.kwargs['jobid'] = task.jobid or self.get_now_string()
        USER_AGENT = settings_dict.pop('USER_AGENT', '')
        # Chrome|iPhone|iPad|Android
        self.kwargs['USER_AGENT'] = dict((v, k) for k, v in UA_DICT.items()).get(USER_AGENT, '')
        for k in ['ROBOTSTXT_OBEY', 'COOKIES_ENABLED']:
            v = settings_dict.pop(k, '')
            self.kwargs[k] = v if v in ['True', 'False'] else ''
        self.kwargs['CONCURRENT_REQUESTS'] = settings_dict.pop('CONCURRENT_REQUESTS', '')
        self.kwargs['DOWNLOAD_DELAY'] = settings_dict.pop('DOWNLOAD_DELAY', '')
        # "-d setting=CLOSESPIDER_TIMEOUT=60\r\n-d setting=CLOSESPIDER_PAGECOUNT=10\r\n-d arg1=val1"
        additional = ''
        # Use sorted() for Python 2
        for k, v in sorted(settings_dict.items()):
            additional += "-d setting=%s=%s\r\n" % (k, v)
        for k, v in sorted(arguments_dict.items()):
            additional += "-d %s=%s\r\n" % (k, v)
        # print(repr(additional))
        self.kwargs['additional'] = additional

        self.kwargs['expand_timer_task'] = True
        self.kwargs['task_id'] = self.task_id
        self.kwargs['name'] = task.name or 'task #%s' % self.task_id
        if not self.kwargs['name'].endswith(' - edit'):
            self.kwargs['name'] += ' - edit'

        self.kwargs['year'] = task.year or '*'
        self.kwargs['month'] = task.month or '*'
        self.kwargs['day'] = task.day or '*'
        self.kwargs['week'] = task.week or '*'
        # To avoid SyntaxError in javascript with Python 2: day_of_week: [u'*'],
        self.kwargs['day_of_week'] = [str(s.strip()) for s in task.day_of_week.split(',')] or ['*']  # 'mon-fri,sun'
        self.kwargs['hour'] = task.hour or '*'
        self.kwargs['minute'] = task.minute or '0'
        self.kwargs['second'] = task.second or '0'

        self.kwargs['start_date'] = task.start_date or ''
        self.kwargs['end_date'] = task.end_date or ''

        if task.timezone:  # To avoid showing 'None' when editing the task
            self.kwargs['timezone'] = task.timezone
        self.kwargs['jitter'] = max(0, task.jitter)
        self.kwargs['misfire_grace_time'] = max(0, task.misfire_grace_time)
        self.kwargs['coalesce'] = task.coalesce if task.coalesce in ['True', 'False'] else 'True'
        self.kwargs['max_instances'] = max(1, task.max_instances)

    def update_kwargs(self):
        self.kwargs.update(dict(
            node=self.node,
            url=self.url,
            url_deploy=url_for('deploy', node=self.node),
            project=self.project,
            version=self.version,
            spider=self.spider,
            # jobid=self.get_now_string(),
            selected_nodes=self.selected_nodes,
            first_selected_node=self.first_selected_node,
            url_servers=url_for('servers', node=self.node, opt='schedule'),
            url_schedule_run=url_for('schedule.run', node=self.node),
            url_schedule_history=url_for('schedule.history'),
            url_listprojects=url_for('api', node=self.node, opt='listprojects'),
            url_listversions=url_for('api', node=self.node, opt='listversions', project='PROJECT_PLACEHOLDER'),
            url_listspiders=url_for('api', node=self.node, opt='listspiders', project='PROJECT_PLACEHOLDER',
                                    version_spider_job='VERSION_PLACEHOLDER'),
            url_schedule_check=url_for('schedule.check', node=self.node)
        ))
        self.kwargs.setdefault('expand_settings_arguments', self.SCHEDULE_EXPAND_SETTINGS_ARGUMENTS)
        self.kwargs.setdefault('jobid', '')
        # self.kwargs.setdefault('UA_DICT', UA_DICT)
        self.kwargs.setdefault('CUSTOM_USER_AGENT', self.SCHEDULE_CUSTOM_USER_AGENT)
        # custom|Chrome|iPhone|iPad|Android
        self.kwargs.setdefault('USER_AGENT', '' if self.SCHEDULE_USER_AGENT is None else self.SCHEDULE_USER_AGENT)
        self.kwargs.setdefault('ROBOTSTXT_OBEY', '' if self.SCHEDULE_ROBOTSTXT_OBEY is None else self.SCHEDULE_ROBOTSTXT_OBEY)
        self.kwargs.setdefault('COOKIES_ENABLED', '' if self.SCHEDULE_COOKIES_ENABLED is None else self.SCHEDULE_COOKIES_ENABLED)
        self.kwargs.setdefault('CONCURRENT_REQUESTS', '' if self.SCHEDULE_CONCURRENT_REQUESTS is None else self.SCHEDULE_CONCURRENT_REQUESTS)
        self.kwargs.setdefault('DOWNLOAD_DELAY', '' if self.SCHEDULE_DOWNLOAD_DELAY is None else self.SCHEDULE_DOWNLOAD_DELAY)
        # additional = "-d setting=CLOSESPIDER_TIMEOUT=60\r\n-d setting=CLOSESPIDER_PAGECOUNT=10\r\n-d arg1=val1"
        self.kwargs.setdefault('additional', self.SCHEDULE_ADDITIONAL)

        self.kwargs.setdefault('expand_timer_task', 'add_task' in request.args)  # '+' button in the TimeTasks page
        self.kwargs.setdefault('task_id', 0)
        self.kwargs['action'] = 'add_fire'
        self.kwargs['trigger'] = 'cron'
        self.kwargs.setdefault('name', '')
        self.kwargs['replace_existing'] = 'True'

        self.kwargs.setdefault('year', '*')
        self.kwargs.setdefault('month', '*')
        self.kwargs.setdefault('day', '*')
        self.kwargs.setdefault('week', '*')
        self.kwargs.setdefault('day_of_week', ['*'])  # 'mon-fri, sun'
        self.kwargs.setdefault('hour', '*')
        self.kwargs.setdefault('minute', '0')
        self.kwargs.setdefault('second', '0')

        self.kwargs.setdefault('start_date', '')
        self.kwargs.setdefault('end_date', '')

        self.kwargs.setdefault('timezone', self.scheduler.timezone)
        self.kwargs.setdefault('jitter', 0)
        self.kwargs.setdefault('misfire_grace_time', 600)
        self.kwargs.setdefault('coalesce', 'True')
        self.kwargs.setdefault('max_instances', 1)


class ScheduleCheckView(BaseView):

    def __init__(self):
        super(ScheduleCheckView, self).__init__()

        self.url = 'http://%s/schedule.json' % self.SCRAPYD_SERVER
        self.template = 'scrapydweb/schedule.html'

        self.filename = ''
        self.data = OrderedDict()
        self.slot = slot

    def dispatch_request(self, **kwargs):
        self.logger.debug('request.form from %s\n%s', request.url, self.json_dumps(request.form))
        self.prepare_data()
        self.update_data_for_timer_task()
        # self.logger.warning(self.json_dumps(self.data))  # TypeError: Object of type datetime is not JSON serializable
        cmd = generate_cmd(self.AUTH, self.url, self.data)
        # '-d' may be in project name, like 'ScrapydWeb-demo'
        cmd = re.sub(r'(curl -u\s+.*?:.*?)\s+(http://)', r'\1 \\\r\n\2', cmd)
        cmd = re.sub(r'\s+-d\s+', ' \\\r\n-d ', cmd)
        cmd = re.sub(r'\s+--data-urlencode\s+', ' \\\r\n--data-urlencode ', cmd)
        return self.json_dumps({'filename': self.filename, 'cmd': cmd}, as_response=True)

    def prepare_data(self):
        for k, d in [('project', 'projectname'), ('_version', self.DEFAULT_LATEST_VERSION),
                     ('spider', 'spidername')]:
            self.data[k] = request.form.get(k, d)
        if self.data['_version'] == self.DEFAULT_LATEST_VERSION:
            self.data.pop('_version')

        jobid = request.form.get('jobid') or self.get_now_string()
        self.data['jobid'] = re.sub(self.LEGAL_NAME_PATTERN, '-', jobid)

        self.data['setting'] = []
        ua = UA_DICT.get(request.form.get('USER_AGENT', ''), '')
        if ua:
            self.data['setting'].append('USER_AGENT=%s' % ua)

        for key in ['ROBOTSTXT_OBEY', 'COOKIES_ENABLED', 'CONCURRENT_REQUESTS', 'DOWNLOAD_DELAY']:
            value = request.form.get(key, '')
            if value:
                self.data['setting'].append("%s=%s" % (key, value))

        additional = request.form.get('additional', '').strip()
        if additional:
            parts = [i.strip() for i in re.split(r'-d\s+', re.sub(r'[\r\n]', ' ', additional)) if i.strip()]
            for part in parts:
                part = re.sub(r'\s*=\s*', '=', part)
                if '=' not in part:
                    continue
                m_setting = re.match(r'setting=([A-Z_]{6,31}=.+)', part)  # 'EDITOR' 'DOWNLOADER_CLIENTCONTEXTFACTORY'
                if m_setting:
                    self.data['setting'].append(m_setting.group(1))
                    continue
                m_arg = re.match(r'([a-zA-Z_][0-9a-zA-Z_]*)=(.+)', part)
                if m_arg and m_arg.group(1) != 'setting':
                    self.data[m_arg.group(1)] = m_arg.group(2)

        self.data['setting'].sort()
        _version = self.data.get('_version', 'default-the-latest-version')
        _filename = '{project}_{version}_{spider}'.format(project=self.data['project'],
                                                          version=_version,
                                                          spider=self.data['spider'])
        self.filename = '%s.pickle' % re.sub(self.LEGAL_NAME_PATTERN, '-', _filename)
        filepath = os.path.join(self.SCHEDULE_PATH, self.filename)
        with io.open(filepath, 'wb') as f:
            f.write(pickle.dumps(self.data))

        self.slot.add_data(self.filename, self.data)

    def get_int_from_form(self, key, default, minimum):
        value = request.form.get(key) or default
        try:
            return max(minimum, int(ceil(float(value))))
        except (TypeError, ValueError) as err:
            self.logger.warning("%s. The value of request.form['%s'] would be set as %s", err, key, default)
            return default

    def update_data_for_timer_task(self):
        if not request.form.get('trigger'):
            return
        # In case passing '-d task_data=xxx' in the additional text box
        self.data['__task_data'] = dict(
            action=request.form.get('action') or 'add_fire',
            task_id=request.form.get('task_id', default=0, type=int),

            # trigger=request.form.get('trigger') or 'cron',
            trigger='cron',
            # id =  # put off in ScheduleRunView.db_insert_task()
            name=request.form.get('name') or None,  # (str) – the description of this job   None
            replace_existing=request.form.get('replace_existing', 'True') == 'True',

            year=request.form.get('year') or '*',  # (int|str) – 4-digit year
            month=request.form.get('month') or '*',  # (int|str) – month (1-12)
            day=request.form.get('day') or '*',  # (int|str) – day of the (1-31)
            week=request.form.get('week') or '*',  # (int|str) – ISO week (1-53)
            # (int|str) – number or name of weekday (0-6 or mon,tue,wed,thu,fri,sat,sun)
            day_of_week=request.form.get('day_of_week') or '*',  # From browser: "*"|"*,mon-fri"|"", May be '0',
            hour=request.form.get('hour') or '*',  # (int|str) – hour (0-23) May be '0'
            minute=request.form.get('minute') or '0',  # (int|str) – minute (0-59)    0
            second=request.form.get('second') or '0',  # (int|str) – second (0-59)    0

            start_date=request.form.get('start_date') or None,  # (datetime|str)    None
            end_date=request.form.get('end_date') or None,  # (datetime|str)        None
            # from tzlocal import get_localzone
            # <DstTzInfo 'Asia/Shanghai' LMT+8:06:00 STD>   +8
            # <DstTzInfo 'US/Eastern' LMT-1 day, 19:04:00 STD>  -5
            timezone=request.form.get('timezone') or None,  # (datetime.tzinfo|str)     defaults to scheduler timezone
            jitter=self.get_int_from_form('jitter', 0, minimum=0),  # (int|None)
            # TypeError: misfire_grace_time must be either None or a positive integer
            # Passing '0' would be saved as None for positive infinity.
            misfire_grace_time=self.get_int_from_form('misfire_grace_time', 600, minimum=0) or None,  # (int)
            coalesce=(request.form.get('coalesce') or 'True') == 'True',  # (bool)
            # TypeError: max_instances must be a positive integer
            max_instances=self.get_int_from_form('max_instances', 1, minimum=1)  # (int)
        )


class ScheduleRunView(BaseView):

    def __init__(self):
        super(ScheduleRunView, self).__init__()

        self.url = ''
        self.template = 'scrapydweb/schedule_results.html'

        self.slot = slot
        self.selected_nodes_amount = 0
        self.selected_nodes = []
        self.first_selected_node = 0
        self.filename = request.form['filename']
        self.data = {}
        self.task_data = {}
        self.task = None
        self.task_id = 0
        self._action = ''
        self.to_update_task = False
        self.add_task_result = False
        self.add_task_flash = ''
        self.add_task_error = ''
        self.add_task_message = ''
        self.js = {}

    def dispatch_request(self, **kwargs):
        self.handle_form()
        self.handle_action()
        self.update_history()
        return self.generate_response()

    def handle_form(self):
        self.selected_nodes_amount = request.form.get('checked_amount', default=0, type=int)
        # With multinodes, would try to Schedule to the first selected node first
        if self.selected_nodes_amount:
            self.selected_nodes = self.get_selected_nodes()
            self.first_selected_node = self.selected_nodes[0]
            self.url = 'http://%s/schedule.json' % self.SCRAPYD_SERVERS[self.first_selected_node - 1]
            # Note that self.first_selected_node != self.node
            self.AUTH = self.SCRAPYD_SERVERS_AUTHS[self.first_selected_node - 1]
        else:
            self.selected_nodes = [self.node]
            self.url = 'http://%s/schedule.json' % self.SCRAPYD_SERVER

        # in handle_action():   self.data.pop('__task_data', {})    self.task_data.pop
        self.data = self.slot.data.get(self.filename, {})
        # self.data = None  # For test only
        if not self.data:
            filepath = os.path.join(self.SCHEDULE_PATH, self.filename)
            with io.open(filepath, 'rb') as f:
                self.data = pickle.loads(f.read())

    def handle_action(self):
        self.logger.debug(self.json_dumps(self.data))
        self.task_data = self.data.pop('__task_data', {})  # Now self.data is clean
        self.logger.debug("task_data: %s", self.task_data)
        if self.task_data:  # For timer task
            self._action = self.task_data.pop('action')  # add|add_fire|add_pause
            self.task_id = self.task_data.pop('task_id')  # 0|positive int from edit button in the Timer Tasks page
            self.to_update_task = self.task_data.pop('replace_existing') and self.task_id  # replace_existing: bool
            self.db_insert_update_task()
            self.add_update_task()
        else:
            self._action = 'run'
            status_code, self.js = self.make_request(self.url, data=self.data, auth=self.AUTH)

    # https://apscheduler.readthedocs.io/en/latest/userguide.html
    # https://apscheduler.readthedocs.io/en/latest/modules/triggers/cron.html#module-apscheduler.triggers.cron
    def db_insert_update_task(self):
        if self.to_update_task:
            self.task = Task.query.get(self.task_id)
            self.logger.debug("Selected %s", self.task)
            self.db_process_task()
            self.task.update_time = datetime.now()
            # Put off in add_update_task()
            # db.session.commit()
            # self.logger.debug("Updated %s", self.task)
        else:
            self.task = Task()
            self.db_process_task()
            db.session.add(self.task)
            db.session.commit()
            self.logger.debug("Inserted %s", self.task)
            self.task_id = self.task.id

    def db_process_task(self):
        data = dict(self.data)  # Used in update_history() and generate_response()

        self.task.project = data.pop('project')
        self.task.version = data.pop('_version', self.DEFAULT_LATEST_VERSION)
        self.task.spider = data.pop('spider')
        self.task.jobid = data.pop('jobid')
        self.task.settings_arguments = self.json_dumps(data, sort_keys=True, indent=None)
        self.task.selected_nodes = str(self.selected_nodes)

        self.task.name = self.task_data['name']
        self.task.trigger = self.task_data['trigger']

        self.task.year = self.task_data['year']
        self.task.month = self.task_data['month']
        self.task.day = self.task_data['day']
        self.task.week = self.task_data['week']
        self.task.day_of_week = self.task_data['day_of_week']
        self.task.hour = self.task_data['hour']
        self.task.minute = self.task_data['minute']
        self.task.second = self.task_data['second']
        self.task.start_date = self.task_data['start_date']
        self.task.end_date = self.task_data['end_date']

        self.task.timezone = self.task_data['timezone']
        self.task.jitter = self.task_data['jitter']
        self.task.misfire_grace_time = self.task_data['misfire_grace_time']
        self.task.coalesce = 'True' if self.task_data['coalesce'] else 'False'  # bool True would be stored as 1
        self.task.max_instances = self.task_data['max_instances']

    # https://apscheduler.readthedocs.io/en/latest/modules/schedulers/base.html#apscheduler.schedulers.base.BaseScheduler.add_job
    def add_update_task(self):
        # class apscheduler.schedulers.base.BaseScheduler
        # def add_job(self, func, trigger=None, args=None, kwargs=None, id=None, name=None,
        #             misfire_grace_time=undefined, coalesce=undefined, max_instances=undefined,
        #             next_run_time=undefined, jobstore='default', executor='default',
        #             replace_existing=False, **trigger_args):
        # TODO: hard coding of url_schedule_task
        # if 'url_schedule_task' not in self.metadata:
        #     url_schedule_task = url_for('schedule.task', node=1)  # /1/schedule/task/
        #     handle_metadata('url_schedule_task', url_schedule_task)
        kwargs = dict(task_id=self.task_id)
        self.task_data['id'] = str(self.task_id)  # TypeError: id must be a nonempty string
        # apscheduler.executors.default: Job "execute_task (trigger: cron[year='*'..." executed successfully
        self.task_data['name'] = self.task_data['name'] or 'task_%s' % self.task_id  # To replace execute_task with name

        # next_run_time (datetime) – when to first run the job, regardless of the trigger
        # (pass None to add the job as paused)
        if self._action == 'add_fire':
            # In case the task fires before db.session.commit()
            if self.to_update_task:
                self.logger.info("Task #%s would be fired right after the apscheduler_job is updated", self.task_id)
            else:
                self.task_data['next_run_time'] = datetime.now()  # datetime.utcnow()
            postfix = "Reload this page several seconds later to check out the execution result. "
        elif self._action == 'add_pause':
            self.task_data['next_run_time'] = None
            postfix = "Click the Paused button to resume it. "
        else:
            postfix = "Click the Running button to pause it. "

        # https://apscheduler.readthedocs.io/en/latest/modules/schedulers/base.html#apscheduler.schedulers.base.BaseScheduler.add_job
        msg = ''
        try:
            # assert False, u"'故意出错'\r\n\"出错\""
            job_instance = self.scheduler.add_job(func=execute_task, args=None, kwargs=kwargs,
                                                  replace_existing=True, **self.task_data)
        except Exception as err:
            # ValueError: Unrecognized expression "10/*" for field "second"
            if self.to_update_task:
                db.session.rollback()
                self.logger.warning("Rollback %s", self.task)
            self.add_task_result = False
            self.add_task_error = str(err)
            msg = traceback.format_exc()
            self.logger.error(msg)
        else:
            # https://apscheduler.readthedocs.io/en/latest/modules/triggers/cron.html#daylight-saving-time-behavior
            # either use a timezone that does not observe DST, for instance UTC
            # https://www.douban.com/note/147740972/
            # F12 Date() Tue Jan 29 2019 13:30:57 GMT+0800 (China Standard Time)
            # https://www.timeanddate.com/time/zones/cst
            # Other time zones named CST: China Standard Time, Cuba Standard Time
            if self.to_update_task:
                db.session.commit()
                self.logger.debug("Updated %s", self.task)
                # In case the task fires before db.session.commit()
                if self._action == 'add_fire':
                    self.logger.info("Modifying next_run_time of updated task #%s to fire it right now", self.task_id)
                    job_instance.modify(next_run_time=datetime.now())
            self.add_task_result = True
            msg = u"{target} task #{task_id} ({task_name}) successfully, next run at {next_run_time}. ".format(
                target="Update" if self.to_update_task else 'Add',
                task_id=self.task_id, task_name=self.task_data['name'],
                next_run_time=job_instance.next_run_time or self.NA)
            self.add_task_flash = msg + postfix
            apscheduler_logger.warning(msg)
            # TypeError: vars() argument must have __dict__ attribute
            # apscheduler_logger.warning(vars(job_instance))
            # pformat({k: getattr(job_instance, k) for k in job_instance.__slots__}, indent=4)
            job_instance_dict = dict(
                id=job_instance.id,
                name=job_instance.name,
                kwargs=job_instance.kwargs,
                misfire_grace_time=job_instance.misfire_grace_time,
                max_instances=job_instance.max_instances,
                trigger=repr(job_instance.trigger),
                next_run_time=repr(job_instance.next_run_time),
            )
            apscheduler_logger.warning("%s job_instance: \n%s", "Updated" if self.to_update_task else 'Added',
                                       self.json_dumps(job_instance_dict))
        finally:
            if 'next_run_time' in self.task_data:  # TypeError: Object of type datetime is not JSON serializable
                self.task_data['next_run_time'] = str(self.task_data['next_run_time'] or self.NA)
            self.add_task_message = (u"{msg}\nkwargs for execute_task():\n{kwargs}\n\n"
                                     u"task_data for scheduler.add_job():\n{task_data}").format(
                msg=msg, kwargs=self.json_dumps(kwargs), task_data=self.json_dumps(self.task_data))
            self.logger.debug(self.add_task_message)

    def update_history(self):
        with io.open(RUN_SPIDER_HISTORY_LOG, 'r+', encoding='utf-8') as f:
            content_backup = f.read()
            f.seek(0)
            content = os.linesep.join([
                '%s %s <%s>' % ('#' * 50, self.get_now_string(True), self._action),
                str([self.SCRAPYD_SERVERS[i - 1] for i in self.selected_nodes]),
                generate_cmd(self.AUTH, self.url, self.data),
                self.add_task_message or self.json_dumps(self.js),
                ''
            ])
            f.write(content)
            f.write(content_backup)

    def generate_response(self):
        if self._action in ['add', 'add_fire', 'add_pause']:
            if self.add_task_result:
                return redirect(url_for('tasks', node=self.node, flash=self.add_task_flash))
            else:
                return render_template(self.template_fail, node=self.node,
                                       alert="Fail to add/edit task with error:",
                                       text=self.add_task_error,
                                       tip=("Check out the HELP section in the Run Spider page, and then "
                                            "go back to the Timer Tasks page to re-edit task #%s. ") % self.task_id,
                                       message=self.add_task_message)
        if self.js['status'] == self.OK:
            if not self.selected_nodes_amount:
                return redirect(url_for('jobs', node=self.node))

            kwargs = dict(
                node=self.node,
                project=self.data['project'],
                version=self.data.get('_version', self.DEFAULT_LATEST_VERSION),
                spider=self.data['spider'],
                selected_nodes=self.selected_nodes,
                first_selected_node=self.first_selected_node,
                js=self.js,
                url_stats_list=[url_for('log', node=node, opt='stats', project=self.data['project'],
                                        spider=self.data['spider'], job=self.data['jobid'])
                                for node in range(1, self.SCRAPYD_SERVERS_AMOUNT + 1)],
                url_xhr=url_for('schedule.xhr', node=self.node, filename=self.filename),
                url_servers=url_for('servers', node=self.node, opt='getreports', project=self.data['project'],
                                    spider=self.data['spider'], version_job=self.data['jobid'])
            )
            return render_template(self.template, **kwargs)
        else:
            if self.selected_nodes_amount > 1:
                alert = ("Multinode schedule terminated, "
                         "since the first selected node returned status: " + self.js['status'])
            else:
                alert = "Fail to schedule, got status: " + self.js['status']

            message = self.js.get('message', '')
            if message:
                self.js['message'] = 'See details below'

            return render_template(self.template_fail, node=self.node,
                                   alert=alert, text=self.json_dumps(self.js), message=message)


class ScheduleXhrView(BaseView):

    def __init__(self):
        super(ScheduleXhrView, self).__init__()

        self.filename = self.view_args['filename']
        self.url = 'http://%s/schedule.json' % self.SCRAPYD_SERVER
        self.slot = slot
        self.data = None

    def dispatch_request(self, **kwargs):
        self.data = self.slot.data.get(self.filename)
        # self.data = None  # For test only
        if not self.data:
            filepath = os.path.join(self.SCHEDULE_PATH, self.filename)
            with io.open(filepath, 'rb') as f:
                self.data = pickle.loads(f.read())

        status_code, js = self.make_request(self.url, data=self.data, auth=self.AUTH)
        return self.json_dumps(js, as_response=True)


class ScheduleTaskView(BaseView):

    def __init__(self):
        super(ScheduleTaskView, self).__init__()

        self.url = 'http://%s/schedule.json' % self.SCRAPYD_SERVER
        self.listjobs_url = 'http://%s/listjobs.json' % self.SCRAPYD_SERVER
        self.task_id = request.form['task_id']
        self.jobid = request.form['jobid']
        self.data = {}

    def dispatch_request(self, **kwargs):
        task = Task.query.get(self.task_id)
        if not task:
            message = "Task #%s not found" % self.task_id
            self.logger.error(message)
            js = dict(url=self.url, auth=self.AUTH, status_code=-1, status=self.ERROR, message=message)
        else:
            self.data['project'] = task.project
            if task.version != self.DEFAULT_LATEST_VERSION:
                self.data['_version'] = task.version
            self.data['spider'] = task.spider
            self.data['jobid'] = self.jobid
            self.data.update(json.loads(task.settings_arguments))
            running_scheduled_count = self.get_number_scheduled_or_running_tasks(task.project, task.spider)
            if running_scheduled_count < task.max_instances:
                status_code, js = self.make_request(self.url, data=self.data, auth=self.AUTH)
            else:
                message = ("The number of scheduled or running tasks of project '%s' and spider '%s' "
                           "has reached the limit %s.") % (task.project, task.spider, task.max_instances)
                self.logger.info(message)
                js = dict(url=self.url, auth=self.AUTH, status_code=-1, status=self.ERROR, message=message)

        return self.json_dumps(js, as_response=True)

    def get_number_scheduled_or_running_tasks(self, project, spider):
        status_code, js = self.make_request(self.listjobs_url, auth=self.AUTH)
        if status_code != 200:
            return None
        count_pending = 0
        count_running = 0
        for pending_job in js.get('pending', []):
            if pending_job['project'] == project and pending_job['spider'] == spider:
                count_pending += 1
        for running_job in js.get('running', []):
            if running_job['project'] == project and running_job['spider'] == spider:
                count_running += 1
        return count_pending + count_running



# coding: utf-8
import io
import json
import os

from flask import url_for

from scrapydweb.utils.check_app_config import check_app_config
from scrapydweb.views.files.log import REPORT_KEYS_SET
from tests.utils import cst, req, replace_file_content, sleep


# http://127.0.0.1:5000/1/log/stats/ScrapydWeb_demo/test/ScrapydWeb_demo/
# id="finish_reason">finished<
# NO refresh_button

# http://127.0.0.1:5000/1/log/stats/ScrapydWeb_demo/test/ScrapydWeb_demo_unfinished/
# id="finish_reason">N/A<
# <a id="refresh_button" class="button danger" href="javascript:location.reload(true);"
# onclick="showLoader();">Click to refresh</a>
# var by = 'ScrapydWeb';
# var click = 'click to hard reparse (SLOW)';
# my$('#refresh_button').innerHTML = "Parsed by " + by +
def test_stats_with_logparser_disabled(app, client):
    kws = dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=cst.DEMO_JOBID)
    req(app, client, view='log', kws=kws,
        ins=["Using local logfile:", 'id="finish_reason">finished<'], nos='refresh_button')

    kws['job'] = cst.DEMO_UNFINISHED_LOG.split('.')[0]
    ins = ['id="finish_reason">N/A<', '<a id="refresh_button"', "var by = 'ScrapydWeb';"]
    req(app, client, view='log', kws=kws, ins=ins)


def test_enable_logparser(app, client):
    def json_loads_from_file(path):
        with io.open(path, 'r', encoding='utf-8') as f:
            return json.loads(f.read())

    # In conftest.py: ENABLE_LOGPARSER=False
    assert not os.path.exists(app.config['STATS_JSON_PATH'])
    assert not os.path.exists(app.config['DEMO_JSON_PATH'])
    app.config['ENABLE_LOGPARSER'] = True
    app.config['ENABLE_MONITOR'] = False

    # ['username:password@127.0.0.1:6800', ]
    app.config['SCRAPYD_SERVERS'] = app.config['_SCRAPYD_SERVERS']
    check_app_config(app.config)

    logparser_pid = app.config['LOGPARSER_PID']
    assert isinstance(logparser_pid, int) and logparser_pid > 0
    assert app.config['POLL_PID'] is None
    req(app, client, view='settings', kws=dict(node=1), ins='logparser_pid: %s' % logparser_pid)

    sleep()

    stats_json = json_loads_from_file(app.config['STATS_JSON_PATH'])
    assert stats_json['logparser_version'] == cst.LOGPARSER_VERSION
    assert cst.DEMO_JOBID in stats_json['datas'][cst.PROJECT][cst.SPIDER]
    demo_json = json_loads_from_file(app.config['DEMO_JSON_PATH'])
    assert demo_json['runtime'] == '0:01:08'
    assert demo_json['finish_reason'] == 'finished'
    assert demo_json['logparser_version'] == cst.LOGPARSER_VERSION


def test_stats_with_logparser_enabled(app, client):
    flash_logparser_version = 'LogParser v%s' % cst.LOGPARSER_VERSION

    kws = dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=cst.DEMO_JOBID)
    req(app, client, view='log', kws=kws,
        ins=[flash_logparser_version, 'id="finish_reason">finished<'], nos='refresh_button')

    # <a class="button danger" href="/1/log/stats/ScrapydWeb_demo/test/ScrapydWeb_demo_unfinished/?realtime=True"
    # onclick="showLoader();">Realtime version</a>
    demo_unfinished_log_without_ext = cst.DEMO_UNFINISHED_LOG.split('.')[0]
    kws['job'] = demo_unfinished_log_without_ext
    with app.test_request_context():
        url_realtime = url_for('log', node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER,
                               job=demo_unfinished_log_without_ext, realtime='True')
    ins = [flash_logparser_version, '<a id="refresh_button"', "javascript:location.reload(true);",
           url_realtime, ">Realtime version</a>", 'id="finish_reason">N/A<', "var by = 'LogParser';"]
    req(app, client, view='log', kws=kws, ins=ins)

    # <a class="button safe" href="/1/log/stats/ScrapydWeb_demo/test/ScrapydWeb_demo_unfinished/"
    # onclick="showLoader();">Cached version</a>
    kws['realtime'] = 'True'
    with app.test_request_context():
        url_cached = url_for('log', node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER,
                             job=demo_unfinished_log_without_ext)
    nos = flash_logparser_version
    ins = [url_cached, ">Cached version</a>", '<a id="refresh_button"', "javascript:location.reload(true);",
           'id="finish_reason">N/A<', "var by = 'ScrapydWeb';"]
    req(app, client, view='log', kws=kws, nos=nos, ins=ins)


# Note that DEMO_JSON is used in test_log.py, avoid removing it
def test_stats_with_file_deleted(app, client):
    tab = ">Log analysis</li>"
    old = '"logparser_version": "%s",' % cst.LOGPARSER_VERSION
    new = '"logparser_version": "0.0.0",'

    def rename(name, restore=False):
        if restore:
            os.rename(name + '.bak', name)
        else:
            os.rename(name, name + '.bak')

    kws = dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=cst.DEMO_JOBID)
    req(app, client, view='log', kws=kws, ins=["Using local stats: LogParser v%s" % cst.LOGPARSER_VERSION, tab])

    # Make IS_LOCAL_SCRAPYD_SERVER False to test request_stats_by_logparser() Pass
    app.config['LOCAL_SCRAPYD_SERVER'] = app.config['SCRAPYD_SERVERS'][-1]
    kws = dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=cst.DEMO_JOBID)
    # LogParser v0.8.1, last updated at xxx, http://127.0.0.1:6800/logs/ScrapydWeb_demo/test/ScrapydWeb_demo.json
    req(app, client, view='log', kws=kws, ins=["LogParser v%s, last updated at" % cst.LOGPARSER_VERSION, tab],
        nos="Using local stats:")
    # Get report
    kws_ = dict(kws)
    kws_['opt'] = 'report'
    jskeys = list(REPORT_KEYS_SET)
    req(app, client, view='log', kws=kws_, jskws=dict(status=cst.OK, from_memory=False), jskeys=jskeys)
    req(app, client, view='log', kws=kws_, jskws=dict(status=cst.OK, from_memory=True), jskeys=jskeys)
    kws_['project'] = cst.FAKE_PROJECT
    req(app, client, view='log', kws=kws_, jskws=dict(status=cst.ERROR))

    app.config['LOCAL_SCRAPYD_SERVER'] = app.config['SCRAPYD_SERVERS'][0]

    # Mismatching logparser_version in ScrapydWeb_demo.json in logs
    replace_file_content(app.config['DEMO_JSON_PATH'], old, new)
    req(app, client, view='log', kws=kws,
        ins=["Mismatching logparser_version 0.0.0 in local stats",
             "pip install --upgrade logparser", "Using local logfile:", tab])
    replace_file_content(app.config['DEMO_JSON_PATH'], new, old)

    # delete ScrapydWeb_demo.json in logs
    # os.remove(app.config['DEMO_JSON_PATH'])
    rename(app.config['DEMO_JSON_PATH'])
    kws = dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=cst.DEMO_JOBID)
    req(app, client, view='log', kws=kws,
        ins=["pip install logparser", "Or wait until LogParser parses the log.", "Using local logfile:", tab])

    app.config['ENABLE_LOGPARSER'] = True  # Test flash content for local scrapyd server
    kws = dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=cst.DEMO_JOBID)
    req(app, client, view='log', kws=kws,
        ins=["got code 404, wait until LogParser parses the log.", tab], nos="pip install logparser")

    # delete ScrapydWeb_demo.log in logs
    rename(app.config['DEMO_LOG_PATH'])
    kws = dict(node=1, opt='utf8', project=cst.PROJECT, spider=cst.SPIDER, job=cst.DEMO_JOBID)
    req(app, client, view='log', kws=kws,
        ins=["fail - ScrapydWeb", "404 - No Such Resource", "Fail to request logfile", "with extensions"])

    kws = dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=cst.DEMO_JOBID)
    req(app, client, view='log', kws=kws, ins=["Using backup stats: LogParser v%s" % cst.LOGPARSER_VERSION, tab])

    # Mismatching logparser_version in ScrapydWeb_demo.json in data/stats
    replace_file_content(app.config['BACKUP_DEMO_JSON_PATH'], old, new)
    req(app, client, view='log', kws=kws,
        ins=["fail - ScrapydWeb", "404 - No Such Resource", "Fail to request logfile", "with extensions",
             "Mismatching logparser_version 0.0.0 in backup stats"])
    replace_file_content(app.config['BACKUP_DEMO_JSON_PATH'], new, old)

    # delete ScrapydWeb_demo.json in data/stats
    rename(app.config['BACKUP_DEMO_JSON_PATH'])
    kws = dict(node=1, opt='stats', project=cst.PROJECT, spider=cst.SPIDER, job=cst.DEMO_JOBID)
    req(app, client, view='log', kws=kws,
        ins=["fail - ScrapydWeb", "404 - No Such Resource", "Fail to request logfile", "with extensions"])

    for filepath in ['DEMO_JSON_PATH', 'DEMO_LOG_PATH', 'BACKUP_DEMO_JSON_PATH']:
        rename(app.config[filepath], restore=True)



# coding: utf-8
import os
import re
import sys


DB_APSCHEDULER = 'scrapydweb_apscheduler'
DB_TIMERTASKS = 'scrapydweb_timertasks'
DB_METADATA = 'scrapydweb_metadata'
DB_JOBS = 'scrapydweb_jobs'
DBS = [DB_APSCHEDULER, DB_TIMERTASKS, DB_METADATA, DB_JOBS]

PATTERN_MYSQL = re.compile(r'mysql://(.+?)(?::(.+?))?@(.+?):(\d+)')
PATTERN_POSTGRESQL = re.compile(r'(?:postgres|postgresql)://(.+?)(?::(.+?))?@(.+?):(\d+)')
PATTERN_SQLITE = re.compile(r'sqlite:///(.+)$')

SCRAPYDWEB_TESTMODE = os.environ.get('SCRAPYDWEB_TESTMODE', 'False').lower() == 'true'


def test_database_url_pattern(database_url):
    m_mysql = PATTERN_MYSQL.match(database_url)
    m_postgres = PATTERN_POSTGRESQL.match(database_url)
    m_sqlite = PATTERN_SQLITE.match(database_url)
    return m_mysql, m_postgres, m_sqlite


def setup_database(database_url, database_path):
    database_url = re.sub(r'\\', '/', database_url)
    database_url = re.sub(r'/$', '', database_url)
    database_path = re.sub(r'\\', '/', database_path)
    database_path = re.sub(r'/$', '', database_path)

    m_mysql, m_postgres, m_sqlite = test_database_url_pattern(database_url)
    if m_mysql:
        setup_mysql(*m_mysql.groups())
    elif m_postgres:
        setup_postgresql(*m_postgres.groups())
    else:
        database_path = m_sqlite.group(1) if m_sqlite else database_path
        database_path = os.path.abspath(database_path)
        database_path = re.sub(r'\\', '/', database_path)
        database_path = re.sub(r'/$', '', database_path)
        if not os.path.isdir(database_path):
            os.mkdir(database_path)

    if m_mysql or m_postgres:
        APSCHEDULER_DATABASE_URI = '/'.join([database_url, DB_APSCHEDULER])
        SQLALCHEMY_DATABASE_URI = '/'.join([database_url, DB_TIMERTASKS])
        SQLALCHEMY_BINDS = {
            'metadata': '/'.join([database_url, DB_METADATA]),
            'jobs': '/'.join([database_url, DB_JOBS])
        }
    else:
        # db names for backward compatibility
        APSCHEDULER_DATABASE_URI = 'sqlite:///' + '/'.join([database_path, 'apscheduler.db'])
        # http://flask-sqlalchemy.pocoo.org/2.3/binds/#binds
        SQLALCHEMY_DATABASE_URI = 'sqlite:///' + '/'.join([database_path, 'timer_tasks.db'])
        SQLALCHEMY_BINDS = {
            'metadata': 'sqlite:///' + '/'.join([database_path, 'metadata.db']),
            'jobs': 'sqlite:///' + '/'.join([database_path, 'jobs.db'])
        }

    if SCRAPYDWEB_TESTMODE:
        print("DATABASE_PATH: %s" % database_path)
        print("APSCHEDULER_DATABASE_URI: %s" % APSCHEDULER_DATABASE_URI)
        print("SQLALCHEMY_DATABASE_URI: %s" % SQLALCHEMY_DATABASE_URI)
        print("SQLALCHEMY_BINDS: %s" % SQLALCHEMY_BINDS)
    return APSCHEDULER_DATABASE_URI, SQLALCHEMY_DATABASE_URI, SQLALCHEMY_BINDS, database_path


def drop_database(cur, dbname):
    sql = "DROP DATABASE %s" % dbname
    print(sql)
    try:
        cur.execute(sql)
    except Exception as err:
        print(err)


def setup_mysql(username, password, host, port):
    """
    ModuleNotFoundError: No module named 'MySQLdb'
    pip install mysqlclient
    Python 2: pip install mysqlclient -> MySQLdb/_mysql.c(29) :
    fatal error C1083: Cannot open include file: 'mysql.h': No such file or directory
    https://stackoverflow.com/questions/51294268/pip-install-mysqlclient-returns-fatal-error-c1083-cannot-open-file-mysql-h
    https://www.lfd.uci.edu/~gohlke/pythonlibs/#mysqlclient
    pip install "path to the downloaded mysqlclient.whl file"
    """
    require_version = '0.9.3'  # Dec 18, 2018
    install_command = "pip install --upgrade pymysql"
    try:
        import pymysql
        assert pymysql.__version__ >= require_version, install_command
    except (ImportError, AssertionError):
        sys.exit("Run command: %s" % install_command)
    else:
        # Run scrapydweb: ModuleNotFoundError: No module named 'MySQLdb'
        pymysql.install_as_MySQLdb()

    conn = pymysql.connect(host=host, port=int(port), user=username, password=password,
                           charset='utf8', cursorclass=pymysql.cursors.DictCursor)
    cur = conn.cursor()
    for dbname in DBS:
        if SCRAPYDWEB_TESTMODE:
            drop_database(cur, dbname)
        # pymysql.err.ProgrammingError: (1007, "Can't create database 'scrapydweb_apscheduler'; database exists")
        # cur.execute("CREATE DATABASE IF NOT EXISTS %s CHARACTER SET 'utf8' COLLATE 'utf8_general_ci'" % dbname)
        try:
            cur.execute("CREATE DATABASE %s CHARACTER SET 'utf8' COLLATE 'utf8_general_ci'" % dbname)
        except Exception as err:
            if 'exists' in str(err):
                pass
            else:
                raise
    cur.close()
    conn.close()


def setup_postgresql(username, password, host, port):
    """
    https://github.com/my8100/notes/blob/master/back_end/the-flask-mega-tutorial.md
    When working with database servers such as MySQL and PostgreSQL,
    you have to create the database in the database server before running upgrade.
    """
    require_version = '2.7.7'  # Jan 23, 2019
    install_command = "pip install --upgrade psycopg2"
    try:
        import psycopg2
        assert psycopg2.__version__ >= require_version, install_command
    except (ImportError, AssertionError):
        sys.exit("Run command: %s" % install_command)

    conn = psycopg2.connect(host=host, port=int(port), user=username, password=password)
    conn.set_isolation_level(0)  # https://wiki.postgresql.org/wiki/Psycopg2_Tutorial
    cur = conn.cursor()
    for dbname in DBS:
        if SCRAPYDWEB_TESTMODE:
            # database "scrapydweb_apscheduler" is being accessed by other users
            # DETAIL:  There is 1 other session using the database.
            # To restart postgres server on Windonws -> win+R: services.msc
            drop_database(cur, dbname)

        # https://www.postgresql.org/docs/9.0/sql-createdatabase.html
        # https://stackoverflow.com/questions/9961795/
        # utf8-postgresql-create-database-like-mysql-including-character-set-encoding-a

        # psycopg2.ProgrammingError: invalid locale name: "en_US.UTF-8"
        # https://stackoverflow.com/questions/40673339/
        # creating-utf-8-database-in-postgresql-on-windows10

        # cur.execute("CREATE DATABASE %s ENCODING 'UTF8' LC_COLLATE 'en-US' LC_CTYPE 'en-US'" % dbname)
        # psycopg2.DataError: new collation (en-US) is incompatible with the collation of the template database
        # (Chinese (Simplified)_People's Republic of China.936)
        # HINT:  Use the same collation as in the template database, or use template0 as template.
        try:
            cur.execute("CREATE DATABASE %s ENCODING 'UTF8' LC_COLLATE 'en_US.UTF-8' LC_CTYPE 'en_US.UTF-8'" % dbname)
        except:
            try:
                cur.execute("CREATE DATABASE %s" % dbname)
            except Exception as err:
                # psycopg2.ProgrammingError: database "scrapydweb_apscheduler" already exists
                if 'exists' in str(err):
                    pass
                else:
                    raise
    cur.close()
    conn.close()



# coding: utf-8
import logging
import os
import re

from flask import current_app as app
from flask import Response, flash, g, request, url_for
from flask.views import View
from logparser import __version__ as LOGPARSER_VERSION
from six import text_type

from ..__version__ import __version__ as SCRAPYDWEB_VERSION
from ..common import (get_now_string, get_response_from_view, handle_metadata,
                      handle_slash, json_dumps, session)
from ..vars import (ALLOWED_SCRAPYD_LOG_EXTENSIONS, APSCHEDULER_DATABASE_URI,
                    DATA_PATH, DEMO_PROJECTS_PATH, DEPLOY_PATH, PARSE_PATH,
                    ALERT_TRIGGER_KEYS, LEGAL_NAME_PATTERN, SCHEDULE_ADDITIONAL,
                    SCHEDULE_PATH, STATE_PAUSED, STATE_RUNNING, STATS_PATH, STRICT_NAME_PATTERN)
from ..utils.scheduler import scheduler


class BaseView(View):
    SCRAPYDWEB_VERSION = SCRAPYDWEB_VERSION
    LOGPARSER_VERSION = LOGPARSER_VERSION

    DEMO_PROJECTS_PATH = DEMO_PROJECTS_PATH
    DEPLOY_PATH = DEPLOY_PATH
    PARSE_PATH = PARSE_PATH
    SCHEDULE_PATH = SCHEDULE_PATH
    STATS_PATH = STATS_PATH

    OK = 'ok'
    ERROR = 'error'
    NA = 'N/A'
    INFO = 'info'
    WARN = 'warning'
    DEFAULT_LATEST_VERSION = 'default: the latest version'
    LEGAL_NAME_PATTERN = LEGAL_NAME_PATTERN
    STRICT_NAME_PATTERN = STRICT_NAME_PATTERN
    ALERT_TRIGGER_KEYS = ALERT_TRIGGER_KEYS

    methods = ['GET', 'POST']

    def __init__(self, *args, **kwargs):
        self.logger = logging.getLogger(self.__class__.__name__)
        # Not in the config file
        self.DEFAULT_SETTINGS_PY_PATH = app.config['DEFAULT_SETTINGS_PY_PATH']
        self.SCRAPYDWEB_SETTINGS_PY_PATH = app.config['SCRAPYDWEB_SETTINGS_PY_PATH']
        self.MAIN_PID = app.config['MAIN_PID']
        self.LOGPARSER_PID = app.config['LOGPARSER_PID']
        self.POLL_PID = app.config['POLL_PID']

        # System
        self.DEBUG = app.config.get('DEBUG', False)
        self.VERBOSE = app.config.get('VERBOSE', False)
        self.DATA_PATH = DATA_PATH
        self.APSCHEDULER_DATABASE_URI = APSCHEDULER_DATABASE_URI
        self.SQLALCHEMY_DATABASE_URI = app.config['SQLALCHEMY_DATABASE_URI']
        self.SQLALCHEMY_BINDS = app.config['SQLALCHEMY_BINDS']

        _level = logging.DEBUG if self.VERBOSE else logging.INFO
        self.logger.setLevel(_level)
        logging.getLogger("requests").setLevel(_level)
        logging.getLogger("urllib3").setLevel(_level)

        # if app.testing:
        self.logger.debug('view_args of %s\n%s', request.url, self.json_dumps(request.view_args))
        if request.args:
            self.logger.debug('request.args of %s\n%s', request.url, self.json_dumps(request.args))
        if request.form:
            self.logger.debug('request.form from %s\n%s', request.url, self.json_dumps(request.form))
        if request.json:
            self.logger.debug('request.json from %s\n%s', request.url, self.json_dumps(request.json))
        if request.files:
            self.logger.debug('request.files from %s\n\n    %s\n', request.url, request.files)

        # ScrapydWeb
        self.SCRAPYDWEB_BIND = app.config.get('SCRAPYDWEB_BIND', '0.0.0.0')
        self.SCRAPYDWEB_PORT = app.config.get('SCRAPYDWEB_PORT', 5000)

        self.ENABLE_AUTH = app.config.get('ENABLE_AUTH', False)
        self.USERNAME = app.config.get('USERNAME', '')
        self.PASSWORD = app.config.get('PASSWORD', '')

        self.ENABLE_HTTPS = app.config.get('ENABLE_HTTPS', False)
        self.CERTIFICATE_FILEPATH = app.config.get('CERTIFICATE_FILEPATH', '')
        self.PRIVATEKEY_FILEPATH = app.config.get('PRIVATEKEY_FILEPATH', '')

        self.URL_SCRAPYDWEB = app.config.get('URL_SCRAPYDWEB', 'http://127.0.0.1:5000')

        # Scrapy
        self.SCRAPY_PROJECTS_DIR = app.config.get('SCRAPY_PROJECTS_DIR', '') or self.DEMO_PROJECTS_PATH

        # Scrapyd
        self.SCRAPYD_SERVERS = app.config.get('SCRAPYD_SERVERS', []) or ['127.0.0.1:6800']
        self.SCRAPYD_SERVERS_AMOUNT = len(self.SCRAPYD_SERVERS)
        self.SCRAPYD_SERVERS_GROUPS = app.config.get('SCRAPYD_SERVERS_GROUPS', []) or ['']
        self.SCRAPYD_SERVERS_AUTHS = app.config.get('SCRAPYD_SERVERS_AUTHS', []) or [None]
        self.SCRAPYD_SERVERS_PUBLIC_URLS = (app.config.get('SCRAPYD_SERVERS_PUBLIC_URLS', None)
                                            or [''] * self.SCRAPYD_SERVERS_AMOUNT)

        self.LOCAL_SCRAPYD_SERVER = app.config.get('LOCAL_SCRAPYD_SERVER', '')
        self.LOCAL_SCRAPYD_LOGS_DIR = app.config.get('LOCAL_SCRAPYD_LOGS_DIR', '')
        self.SCRAPYD_LOG_EXTENSIONS = (app.config.get('SCRAPYD_LOG_EXTENSIONS', [])
                                       or ALLOWED_SCRAPYD_LOG_EXTENSIONS)

        # LogParser
        self.ENABLE_LOGPARSER = app.config.get('ENABLE_LOGPARSER', False)
        self.BACKUP_STATS_JSON_FILE = app.config.get('BACKUP_STATS_JSON_FILE', True)

        # Timer Tasks
        self.scheduler = scheduler
        self.JOBS_SNAPSHOT_INTERVAL = app.config.get('JOBS_SNAPSHOT_INTERVAL', 300)

        # Run Spider
        self.SCHEDULE_EXPAND_SETTINGS_ARGUMENTS = app.config.get('SCHEDULE_EXPAND_SETTINGS_ARGUMENTS', False)
        self.SCHEDULE_CUSTOM_USER_AGENT = app.config.get('SCHEDULE_CUSTOM_USER_AGENT', 'Mozilla/5.0')
        self.SCHEDULE_USER_AGENT = app.config.get('SCHEDULE_USER_AGENT', None)
        self.SCHEDULE_ROBOTSTXT_OBEY = app.config.get('SCHEDULE_ROBOTSTXT_OBEY', None)
        self.SCHEDULE_COOKIES_ENABLED = app.config.get('SCHEDULE_COOKIES_ENABLED', None)
        self.SCHEDULE_CONCURRENT_REQUESTS = app.config.get('SCHEDULE_CONCURRENT_REQUESTS', None)
        self.SCHEDULE_DOWNLOAD_DELAY = app.config.get('SCHEDULE_DOWNLOAD_DELAY', None)
        self.SCHEDULE_ADDITIONAL = app.config.get('SCHEDULE_ADDITIONAL', SCHEDULE_ADDITIONAL)

        # Page Display
        self.SHOW_SCRAPYD_ITEMS = app.config.get('SHOW_SCRAPYD_ITEMS', True)
        self.SHOW_JOBS_JOB_COLUMN = app.config.get('SHOW_JOBS_JOB_COLUMN', False)
        self.JOBS_FINISHED_JOBS_LIMIT = app.config.get('JOBS_FINISHED_JOBS_LIMIT', 0)
        self.JOBS_RELOAD_INTERVAL = app.config.get('JOBS_RELOAD_INTERVAL', 300)
        self.DAEMONSTATUS_REFRESH_INTERVAL = app.config.get('DAEMONSTATUS_REFRESH_INTERVAL', 10)

        # Send text
        self.SLACK_TOKEN = app.config.get('SLACK_TOKEN', '')
        self.SLACK_CHANNEL = app.config.get('SLACK_CHANNEL', '') or 'general'
        self.TELEGRAM_TOKEN = app.config.get('TELEGRAM_TOKEN', '')
        self.TELEGRAM_CHAT_ID = app.config.get('TELEGRAM_CHAT_ID', 0)
        self.EMAIL_SUBJECT = app.config.get('EMAIL_SUBJECT', '') or 'Email from #scrapydweb'

        # Monitor & Alert
        self.ENABLE_MONITOR = app.config.get('ENABLE_MONITOR', False)
        self.ENABLE_SLACK_ALERT = app.config.get('ENABLE_SLACK_ALERT', False)
        self.ENABLE_TELEGRAM_ALERT = app.config.get('ENABLE_TELEGRAM_ALERT', False)
        self.ENABLE_EMAIL_ALERT = app.config.get('ENABLE_EMAIL_ALERT', False)

        self.EMAIL_SENDER = app.config.get('EMAIL_SENDER', '')
        self.EMAIL_RECIPIENTS = app.config.get('EMAIL_RECIPIENTS', [])
        self.EMAIL_USERNAME = app.config.get('EMAIL_USERNAME', '') or self.EMAIL_SENDER
        self.EMAIL_PASSWORD = app.config.get('EMAIL_PASSWORD', '')

        self.SMTP_SERVER = app.config.get('SMTP_SERVER', '')
        self.SMTP_PORT = app.config.get('SMTP_PORT', 0)
        self.SMTP_OVER_SSL = app.config.get('SMTP_OVER_SSL', False)
        self.SMTP_CONNECTION_TIMEOUT = app.config.get('SMTP_CONNECTION_TIMEOUT', 30)

        self.EMAIL_KWARGS = dict(
            email_username=self.EMAIL_USERNAME,
            email_password=self.EMAIL_PASSWORD,
            email_sender=self.EMAIL_SENDER,
            email_recipients=self.EMAIL_RECIPIENTS,
            smtp_server=self.SMTP_SERVER,
            smtp_port=self.SMTP_PORT,
            smtp_over_ssl=self.SMTP_OVER_SSL,
            smtp_connection_timeout=self.SMTP_CONNECTION_TIMEOUT,
            subject='subject',
            content='content'
        )

        self.POLL_ROUND_INTERVAL = app.config.get('POLL_ROUND_INTERVAL', 300)
        self.POLL_REQUEST_INTERVAL = app.config.get('POLL_REQUEST_INTERVAL', 10)
        self.ALERT_WORKING_DAYS = app.config.get('ALERT_WORKING_DAYS', [])
        self.ALERT_WORKING_HOURS = app.config.get('ALERT_WORKING_HOURS', [])
        self.ON_JOB_RUNNING_INTERVAL = app.config.get('ON_JOB_RUNNING_INTERVAL', 0)
        self.ON_JOB_FINISHED = app.config.get('ON_JOB_FINISHED', False)
        # ['CRITICAL', 'ERROR', 'WARNING', 'REDIRECT', 'RETRY', 'IGNORE']
        for key in self.ALERT_TRIGGER_KEYS:
            setattr(self, 'LOG_%s_THRESHOLD' % key, app.config.get('LOG_%s_THRESHOLD' % key, 0))
            setattr(self, 'LOG_%s_TRIGGER_STOP' % key, app.config.get('LOG_%s_TRIGGER_STOP' % key, False))
            setattr(self, 'LOG_%s_TRIGGER_FORCESTOP' % key, app.config.get('LOG_%s_TRIGGER_FORCESTOP' % key, False))

        # Other attributes not from config
        self.view_args = request.view_args
        self.node = self.view_args['node']
        assert 0 < self.node <= self.SCRAPYD_SERVERS_AMOUNT, \
            'node index error: %s, which should be between 1 and %s' % (self.node, self.SCRAPYD_SERVERS_AMOUNT)
        self.SCRAPYD_SERVER = self.SCRAPYD_SERVERS[self.node - 1]
        self.IS_LOCAL_SCRAPYD_SERVER = self.SCRAPYD_SERVER == self.LOCAL_SCRAPYD_SERVER
        self.GROUP = self.SCRAPYD_SERVERS_GROUPS[self.node - 1]
        self.AUTH = self.SCRAPYD_SERVERS_AUTHS[self.node - 1]
        self.SCRAPYD_SERVER_PUBLIC_URL = self.SCRAPYD_SERVERS_PUBLIC_URLS[self.node - 1]

        ua = request.headers.get('User-Agent', '')
        m_mobile = re.search(r'Android|webOS|iPad|iPhone|iPod|BlackBerry|IEMobile|Opera Mini', ua, re.I)
        self.IS_MOBILE = True if m_mobile else False

        m_ipad = re.search(r'iPad', ua, re.I)
        self.IS_IPAD = True if m_ipad else False

        # http://werkzeug.pocoo.org/docs/0.14/utils/#module-werkzeug.useragents
        # /site-packages/werkzeug/useragents.py
        browser = request.user_agent.browser or ''  # lib requests GET: None
        m_edge = re.search(r'Edge', ua, re.I)
        self.IS_IE_EDGE = True if (browser == 'msie' or m_edge) else False

        self.USE_MOBILEUI = request.args.get('ui', '') == 'mobile'
        self.UI = 'mobile' if self.USE_MOBILEUI else None
        self.GET = request.method == 'GET'
        self.POST = request.method == 'POST'

        self.FEATURES = ''
        self.FEATURES += 'A' if self.ENABLE_AUTH else '-'
        self.FEATURES += 'D' if handle_metadata().get('jobs_style') == 'database' else 'C'
        self.FEATURES += 'd' if self.SCRAPY_PROJECTS_DIR != self.DEMO_PROJECTS_PATH else '-'
        self.FEATURES += 'L' if self.ENABLE_LOGPARSER else '-'
        self.FEATURES += 'Sl' if self.ENABLE_SLACK_ALERT else '-'
        self.FEATURES += 'Tg' if self.ENABLE_TELEGRAM_ALERT else '-'
        self.FEATURES += 'Em' if self.ENABLE_EMAIL_ALERT else '-'
        self.FEATURES += 'P' if self.IS_MOBILE else '-'
        self.FEATURES += 'M' if self.USE_MOBILEUI else '-'
        self.FEATURES += 'S' if self.ENABLE_HTTPS else '-'
        self.any_running_apscheduler_jobs = any(job.next_run_time
                                                for job in self.scheduler.get_jobs(jobstore='default'))
        if self.scheduler.state == STATE_PAUSED:
            self.FEATURES += '-'
        elif self.any_running_apscheduler_jobs:
            self.FEATURES += 'T'
        else:
            self.FEATURES += 't'
        if not self.SQLALCHEMY_DATABASE_URI.startswith('sqlite'):
            self.FEATURES += self.SQLALCHEMY_DATABASE_URI[:3]

        self.template_fail = 'scrapydweb/fail_mobileui.html' if self.USE_MOBILEUI else 'scrapydweb/fail.html'
        self.update_g()

    @staticmethod
    def get_job_without_ext(job):
        if job.endswith('.tar.gz'):
            return job[:-len('.tar.gz')]
        else:
            return os.path.splitext(job)[0]  # '1.1.log' => ('1.1', '.log')

    @staticmethod
    def get_now_string(allow_space=False):
        return get_now_string(allow_space=allow_space)

    def get_response_from_view(self, url, data=None, as_json=False):
        auth = (self.USERNAME, self.PASSWORD) if self.ENABLE_AUTH else None
        return get_response_from_view(url, auth=auth, data=data, as_json=as_json)

    def get_selected_nodes(self):
        selected_nodes = []
        for n in range(1, self.SCRAPYD_SERVERS_AMOUNT + 1):
            if request.form.get(str(n)) == 'on':
                selected_nodes.append(n)
        return selected_nodes

    @staticmethod
    def handle_slash(string):
        return handle_slash(string)

    @staticmethod
    def json_dumps(obj, sort_keys=True, indent=4, ensure_ascii=False, as_response=False):
        # flask.jsonify
        # https://flask.palletsprojects.com/en/1.1.x/config/#JSONIFY_MIMETYPE
        # https://stackoverflow.com/questions/11773348/python-flask-how-to-set-content-type
        # https://stackoverflow.com/questions/9254891/what-does-content-type-application-json-charset-utf-8-really-mean
        js = json_dumps(obj, sort_keys=sort_keys, indent=indent, ensure_ascii=ensure_ascii)
        if as_response:
            # Content-Type: application/json
            return Response(js, mimetype='application/json')
        else:
            return js

    @staticmethod
    def remove_microsecond(dt):
        return str(dt)[:19]

    def make_request(self, url, data=None, auth=None, as_json=True, dumps_json=True, check_status=True, timeout=60):
        """
        :param url: url to make request
        :param data: None or a dict object to post
        :param auth: None or (username, password) for basic auth
        :param as_json: return a dict object if set True, else text
        :param dumps_json: whether to dumps the json response when as_json is set to True
        :param check_status: whether to log error when status != 'ok'
        :param timeout: timeout when making request, in seconds
        """
        try:
            if 'addversion.json' in url and data:
                self.logger.debug(">>>>> POST %s", url)
                self.logger.debug(self.json_dumps(dict(project=data['project'], version=data['version'],
                                                  egg="%s bytes binary egg file" % len(data['egg']))))
            else:
                self.logger.debug(">>>>> %s %s", 'POST' if data else 'GET', url)
                if data:
                    self.logger.debug("POST data: %s", self.json_dumps(data))

            if data:
                r = session.post(url, data=data, auth=auth, timeout=timeout)
            else:
                r = session.get(url, auth=auth, timeout=timeout)
            r.encoding = 'utf-8'
        except Exception as err:
            # self.logger.error('!!!!! %s %s' % (err.__class__.__name__, err))
            self.logger.error("!!!!! error with %s: %s", url, err)
            if as_json:
                r_json = dict(url=url, auth=auth, status_code=-1, status=self.ERROR,
                              message=str(err), when=self.get_now_string(True))
                return -1, r_json
            else:
                return -1, str(err)
        else:
            if as_json:
                r_json = {}
                try:
                    # listprojects would get 502 html when Scrapyd server reboots
                    r_json = r.json()  # PY3: json.decoder.JSONDecodeError  PY2: exceptions.ValueError
                except ValueError as err:  # issubclass(JSONDecodeError, ValueError)
                    self.logger.error("Fail to decode json from %s: %s", url, err)
                    r_json = dict(status=self.ERROR, message=r.text)
                finally:
                    # Scrapyd in Python2: Traceback (most recent call last):\\n
                    # Scrapyd in Python3: Traceback (most recent call last):\r\n
                    message = r_json.get('message', '')
                    if message and not isinstance(message, dict):
                        r_json['message'] = re.sub(r'\\n', '\n', message)
                    r_json.update(dict(url=url, auth=auth, status_code=r.status_code, when=self.get_now_string(True)))
                    status = r_json.setdefault('status', self.NA)
                    if r.status_code != 200 or (check_status and status != self.OK):
                        self.logger.error("!!!!! (%s) %s: %s", r.status_code, status, url)
                    else:
                        self.logger.debug("<<<<< (%s) %s: %s", r.status_code, status, url)
                    if dumps_json:
                        self.logger.debug("Got json from %s: %s", url, self.json_dumps(r_json))
                    else:
                        self.logger.debug("Got keys from (%s) %s %s: %s",
                                          r_json.get('status_code'), r_json.get('status'), url, r_json.keys())

                    return r.status_code, r_json
            else:
                if r.status_code == 200:
                    _text = r.text[:100] + '......' + r.text[-100:] if len(r.text) > 200 else r.text
                    self.logger.debug("<<<<< (%s) %s\n%s", r.status_code, url, repr(_text))
                else:
                    self.logger.error("!!!!! (%s) %s\n%s", r.status_code, url, r.text)

                return r.status_code, r.text

    def update_g(self):
        # g lifetime: every single request
        # Note that use inject_variable() in View class would cause memory leak, issue #14
        g.IS_MOBILE = self.IS_MOBILE
        g.url_jobs_list = [url_for('jobs', node=node, ui=self.UI)
                           for node in range(1, self.SCRAPYD_SERVERS_AMOUNT + 1)]
        g.multinode = ('<label title="multinode">'
                       '<svg class="icon" aria-hidden="true"><use xlink:href="#icon-servers"></use></svg>'
                       '</label>')
        # For base.html
        if not self.USE_MOBILEUI:
            g.url_daemonstatus = url_for('api', node=self.node, opt='daemonstatus')
            g.url_menu_servers = url_for('servers', node=self.node)
            g.url_menu_jobs = url_for('jobs', node=self.node)
            g.url_menu_nodereports = url_for('nodereports', node=self.node)
            g.url_menu_clusterreports = url_for('clusterreports', node=self.node)
            g.url_menu_tasks = url_for('tasks', node=self.node)
            g.url_menu_deploy = url_for('deploy', node=self.node)
            g.url_menu_schedule = url_for('schedule', node=self.node)
            g.url_menu_projects = url_for('projects', node=self.node)
            g.url_menu_logs = url_for('logs', node=self.node)
            g.url_menu_items = url_for('items', node=self.node)
            g.url_menu_sendtext = url_for('sendtext', node=self.node)
            g.url_menu_parse = url_for('parse.upload', node=self.node)
            g.url_menu_settings = url_for('settings', node=self.node)
            g.url_menu_mobileui = url_for('index', node=self.node, ui='mobile')
            g.scheduler_state_paused = self.scheduler.state == STATE_PAUSED and self.any_running_apscheduler_jobs
            g.scheduler_state_running = self.scheduler.state == STATE_RUNNING and self.any_running_apscheduler_jobs

    # Issue#48 [PY2] UnicodeDecodeError raised when there are some files with illegal filenames in `SCRAPY_PROJECTS_DIR`
    # https://stackoverflow.com/questions/21772271/unicodedecodeerror-when-performing-os-walk
    # https://xuanwo.io/2018/04/01/python-os-walk/
    # Tested in Ubuntu:
    # touch $(echo -e "\x8b\x8bFile")
    # mkdir $(echo -e "\x8b\x8bFolder")
    def safe_walk(self, top, topdown=True, onerror=None, followlinks=False):
        islink, join, isdir = os.path.islink, os.path.join, os.path.isdir

        # touch $(echo -e "\x8b\x8bThis is a bad filename")
        # ('top: ', u'/home/username/download/scrapydweb/scrapydweb/data/demo_projects/ScrapydWeb_demo')
        # ('names: ', ['\x8b\x8bThis', u'ScrapydWeb_demo', u'filename', u'scrapy.cfg', u'a', u'is', u'bad'])
        try:
            names = os.listdir(top)
        except OSError as err:
            if onerror is not None:
                onerror(err)
            return

        new_names = []
        for name in names:
            if isinstance(name, text_type):
                new_names.append(name)
            else:
                msg = "Ignore non-unicode filename %s in %s" % (repr(name), top)
                self.logger.error(msg)
                flash(msg, self.WARN)
        names = new_names

        dirs, nondirs = [], []
        for name in names:
            if isdir(join(top, name)):
                dirs.append(name)
            else:
                nondirs.append(name)

        if topdown:
            yield top, dirs, nondirs
        for name in dirs:
            new_path = join(top, name)
            if followlinks or not islink(new_path):
                for x in self.safe_walk(new_path, topdown, onerror, followlinks):
                    yield x
        if not topdown:
            yield top, dirs, nondirs


class MetadataView(BaseView):

    def __init__(self):
        super(MetadataView, self).__init__()

    def dispatch_request(self, **kwargs):
        return self.json_dumps(handle_metadata(), as_response=True)



# coding: utf-8
import atexit
from ctypes import cdll
import logging
import os
import platform
import signal
from subprocess import Popen
import sys

from ..common import json_dumps


logger = logging.getLogger(__name__)

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


# https://stackoverflow.com/a/13256908/10517783
# https://stackoverflow.com/a/23587108/10517783
# http://evans.io/legacy/posts/killing-child-processes-on-parent-exit-prctl/
class PrCtlError(Exception):
    pass


def on_parent_exit(signame):
    """
    Return a function to be run in a child process which will trigger
    SIGNAME to be sent when the parent process dies
    """
    # On Windows, signal() can only be called with SIGABRT, SIGFPE, SIGILL, SIGINT, SIGSEGV, or SIGTERM.
    signum = getattr(signal, signame)  # SIGTERM 15  SIGKILL 9

    def set_parent_exit_signal():
        # Constant taken from http://linux.die.net/include/linux/prctl.h
        PR_SET_PDEATHSIG = 1
        # http://linux.die.net/man/2/prctl
        result = cdll['libc.so.6'].prctl(PR_SET_PDEATHSIG, signum)
        if result != 0:
            raise PrCtlError('prctl failed with error code %s' % result)

    return set_parent_exit_signal


# https://stackoverflow.com/a/19448255/10517783
def kill_child(proc, title=''):
    proc.kill()
    # A None value indicates that the process has not terminated yet.
    # A negative value -N indicates that the child was terminated by signal N (Unix only).
    logger.warning('%s subprocess (pid: %s) killed with returncode: %s', title, proc.pid, proc.wait())


def init_logparser(config):
    logparser_subprocess = start_logparser(config)
    logparser_pid = logparser_subprocess.pid
    logger.info("Running LogParser in the background with pid: %s", logparser_pid)
    atexit.register(kill_child, logparser_subprocess, 'LogParser')
    return logparser_pid


def start_logparser(config):
    args = [
        sys.executable,
        '-m',
        'logparser.run',
        '-dir',
        config['LOCAL_SCRAPYD_LOGS_DIR'],
        '--main_pid',
        str(config['MAIN_PID']),
    ]

    if platform.system() == 'Linux':
        kwargs = dict(preexec_fn=on_parent_exit('SIGKILL'))  # 'SIGTERM' 'SIGKILL'
        try:
            logparser_subprocess = Popen(args, **kwargs)
        except Exception as err:
            logger.error(err)
            logparser_subprocess = Popen(args)
    else:
        logparser_subprocess = Popen(args)

    return logparser_subprocess


def init_poll(config):
    poll_subprocess = start_poll(config)
    poll_pid = poll_subprocess.pid
    logger.info("Start polling job stats for monitor & alert in the background with pid: %s", poll_pid)
    atexit.register(kill_child, poll_subprocess, 'Poll')
    return poll_pid


def start_poll(config):
    args = [
        sys.executable,
        os.path.join(CURRENT_DIR, 'poll.py'),

        config['URL_SCRAPYDWEB'],
        config.get('USERNAME', '') if config.get('ENABLE_AUTH', False) else '',
        config.get('PASSWORD', '') if config.get('ENABLE_AUTH', False) else '',
        json_dumps(config.get('SCRAPYD_SERVERS', ['127.0.0.1'])),
        json_dumps(config.get('SCRAPYD_SERVERS_AUTHS', [None])),
        str(config.get('POLL_ROUND_INTERVAL', 300)),
        str(config.get('POLL_REQUEST_INTERVAL', 10)),
        str(config['MAIN_PID']),
        str(config.get('VERBOSE', False))
    ]

    # 'Windows':
    # AttributeError: module 'signal' has no attribute 'SIGKILL'
    # ValueError: preexec_fn is not supported on Windows platforms
    # macOS('Darwin'):
    # subprocess.SubprocessError: Exception occurred in preexec_fn.
    # OSError: dlopen(libc.so.6, 6): image not found
    if platform.system() == 'Linux':
        kwargs = dict(preexec_fn=on_parent_exit('SIGKILL'))  # 'SIGTERM' 'SIGKILL'
        try:
            poll_subprocess = Popen(args, **kwargs)
        except Exception as err:
            logger.error(err)
            poll_subprocess = Popen(args)
    else:
        poll_subprocess = Popen(args)

    return poll_subprocess



# coding: utf-8
import json
import os
import re
import time
import traceback

from flask import current_app as app
from flask import Response
import requests
from requests.adapters import HTTPAdapter
from w3lib.http import basic_auth_header

from .__version__ import __version__
from .models import Metadata, db


session = requests.Session()
session.mount('http://', HTTPAdapter(pool_connections=1000, pool_maxsize=1000))
session.mount('https://', HTTPAdapter(pool_connections=1000, pool_maxsize=1000))


# http://flask.pocoo.org/snippets/category/authentication/
def authenticate():
    """Sends a 401 response that enables basic auth"""
    return Response("<script>alert('Fail to login: basic auth for ScrapydWeb has been enabled');</script>",
                    401, {'WWW-Authenticate': 'Basic realm="ScrapydWeb Basic Auth Required"'})


def find_scrapydweb_settings_py(filename, path, prevpath=None):
    if path == prevpath:
        return ''
    path = os.path.abspath(path)
    cfgfile = os.path.join(path, filename)
    if os.path.exists(cfgfile):
        return cfgfile
    # In vars.py, try to import module scrapydweb_settings_vN in cwd only
    # return find_scrapydweb_settings_py(filename, os.path.dirname(path), path)


def get_now_string(allow_space=False):
    if allow_space:
        return time.strftime('%Y-%m-%d %H:%M:%S')
    else:
        return time.strftime('%Y-%m-%dT%H_%M_%S')


def get_response_from_view(url, auth=None, data=None, as_json=False):
    # https://stackoverflow.com/a/21342070/10517783  How do I call one Flask view from another one?
    # https://stackoverflow.com/a/30250045/10517783
    # python - Flask test_client() doesn't have request.authorization with pytest
    client = app.test_client()
    if auth is not None:
        headers = {'Authorization': basic_auth_header(*auth)}
    else:
        headers = {}
    if data is not None:
        response = client.post(url, headers=headers, data=data, content_type='multipart/form-data')
        # response = client.post(url, headers=headers, data=data, content_type='application/json')
    else:
        response = client.get(url, headers=headers)

    text = response.get_data(as_text=True)
    if as_json:
        # e.g. when used in schedule_task()
        # 'node index error: %s, which should be between 1 and %s' % (self.node, self.SCRAPYD_SERVERS_AMOUNT)
        # json.decoder.JSONDecodeError: Expecting value: line 1 column 1 (char 0)
        try:
            return json.loads(text)
        except ValueError:
            # '_status': '500 INTERNAL SERVER ERROR',
            # '_status_code': 500,
            # See 500.html
            # <tr><th>Error</th><td>node index error: 2, which should be between 1 and 1</td></tr>
            # <pre>Traceback...AssertionError: node index error: 2, which should be between 1 and 1 </pre>
            m = re.search(r'<tr><th>Error</th><td>(.+?)</td></tr>', text, re.S)
            message = m.group(1) if m else text
            return dict(status_code=getattr(response, '_status_code', 500), status='error', message=message)
    else:
        return text


def handle_metadata(key=None, value=None):
    with db.app.app_context():
        metadata = Metadata.query.filter_by(version=__version__).first()
        if key is None:
            # '_sa_instance_state': <sqlalchemy.orm.state.InstanceState object at 0x0000000005194080>,
            return dict((k, v) for (k, v) in metadata.__dict__.items() if not k.startswith('_')) if metadata else {}
        else:
            try:
                setattr(metadata, key, value)
                db.session.commit()
            except:
                print(traceback.format_exc())
                db.session.rollback()


def handle_slash(string):
    if not string:
        return string
    else:
        return string.replace('\\', '/')


def json_dumps(obj, sort_keys=True, indent=4, ensure_ascii=False):
    return json.dumps(obj, sort_keys=sort_keys, indent=indent, ensure_ascii=ensure_ascii)
